# -*- coding: utf-8 -*-

from __future__ import annotations

import re
from random import choices
from string import ascii_letters, digits, punctuation
from typing import Dict, List, Iterator, ByteString


def random_string(vocabulary: str = None, length: int = 20, exclude: str = None) -> str:
    """ Create random string """

    if not vocabulary:
        vocabulary = ascii_letters + digits + punctuation

    choices_ = []
    while len(choices_) < length:
        char = choices(vocabulary)[0]
        if not exclude or (exclude and char not in exclude):
            choices_.append(char)

    return "".join(choices_)


def get_batches(list_: List, n: int) -> Iterator:
    """
    Divide a list into chunks...

    :param list_: List to divide.
    :param n: Number of elements by chunk.
    :return:
    """

    for i in range(0, len(list_), n):
        yield list_[i:i + n]


def remove_attributes(record: Dict, attrs: List[str]) -> None:
    """ Remove attributes from a Dict """

    for key in attrs:
        if key in record:
            del record[key]


def rename_attributes(record: Dict, mapper: Dict[str, str]) -> None:
    """ Rename the object attributes using the column_mapper """

    _items = list(record.items())
    for key, value in _items:
        if key in mapper:
            record[mapper[key]] = record[key]
            del record[key]


def add_attributes(record: Dict, expected_attrs: List[str]):
    """
    The function add to the dictionary the list of
    attributes (if not exists) expected
    using None...

    :param record: Dictionary to update.
    :param expected_attrs: Attributes to add.
    :return:
    """

    for key in expected_attrs:
        if key not in record:
            record[key] = None


def convert_data_type(data: Dict, columns_type_mapper: Dict) -> None:
    """
    Update (cast) the value depending on the specified type...

    :param data: Dictionary to update.
    :param columns_type_mapper: Dictionary that specify the data types.
    """

    if not columns_type_mapper:
        return

    for key, type_ in columns_type_mapper.items():
        if key in data and data[key] is not None:
            value = data[key].strip() if type_ == "float" else data[key]
            data.update({
                key: eval(f"{type_}({value})")
            })


def to_snake_case(string: str) -> str:
    return "".join(
        ["_" + i.lower() if i.isupper() else i for i in string]
    ).lstrip("_")


def flatten_json(data: Dict, flatten_sublist: bool = False):
    """
    Utility function for flattening dictionary objects...

    :param data: Object to flatten.
    :param flatten_sublist: Set as True if you want to flatten sublist objects.

    :return: Flatten data (dictionary).
    """

    res = {}

    def flatten(x, name: str = ""):
        if isinstance(x, dict):
            for a in x:
                flatten(x[a], name + a + "_")

        elif flatten_sublist and isinstance(x, list):
            i = 0
            for a in x:
                flatten(a, name + str(i) + "_")
                i += 1

        else:
            res[name[:-1]] = x

    flatten(data)
    return res


def to_one_line(multiline_string: str) -> str:
    """ It converts a multiline string to a single line one """

    return re.sub(
        pattern=r"\s+", repl=" ",
        string=multiline_string.replace("\n", "")
    ).strip()


def bytes_to_str(data: List | Dict | ByteString, encoding: str = "utf-8"):
    """
    Convert bytes to str in keys and values...
    Example:
        {b"key": b"value"})         -> {"key": "value"}
        {b"key": [1, 2, b"3"]}      -> {"key": [1, 2, "3"]}
        [1, 2, b"3", {b"a": b"a"}]  -> [1, 2, "3", {"a": "a"}]
        b"test")                    -> "test"
    """

    def convert(obj_, encoding_, result=None):
        if not result:
            result = {}

        if isinstance(obj_, dict):
            dic_ = {}
            for key in obj_:
                previous_key = key
                if isinstance(key, bytes):
                    key = key.decode()

                dic_[key] = convert(obj_[previous_key], encoding_, dic_)

            return dic_

        elif isinstance(obj_, list):
            return [convert(x, encoding_, result or {}) for x in obj_]

        else:
            return obj_.decode() if isinstance(obj_, bytes) else obj_

    return convert(data, encoding)



# -*- coding: utf-8 -*-

import csv

from io import StringIO
from typing import Dict, Iterator, List


def records_from_csv(
        path: str, file_args: Dict = None,
        reader_args: Dict = None) -> Iterator[Dict]:

    """ Returns records from a CSV file """

    file_args = file_args or {}
    reader_args = reader_args or {}

    with open(path, **file_args) as f_:
        reader = csv.DictReader(f_, **reader_args)
        for rec in reader:
            yield rec


def records_from_buffer(
        data: str, newline: str = None, **kwargs) -> Iterator[Dict]:

    """ Returns records from file-like object """

    dialect = kwargs.pop("dialect", "excel")
    quoting = kwargs.pop("quoting", csv.QUOTE_NONNUMERIC)

    reader = csv.DictReader(
        StringIO(data, newline=newline),
        dialect=dialect, quoting=quoting,
        **kwargs)

    for rec in reader:
        yield rec


def records_to_buffer(
        records: List[Dict], columns: List[str] = None,
        **kwargs) -> StringIO:

    """ Save records from file-like object """

    buffer = StringIO()
    dialect = kwargs.pop("dialect", "excel")
    quoting = kwargs.pop("quoting", csv.QUOTE_NONNUMERIC)
    columns = columns or records[0].keys()

    csv_writer = csv.DictWriter(
        buffer, fieldnames=columns,
        dialect=dialect, quoting=quoting,
        **kwargs)

    csv_writer.writeheader()
    csv_writer.writerows(records)
    return buffer



# -*- coding: utf-8 -*-

from functools import update_wrapper, wraps
from typing import Callable


class CountCalls:
    """ Decorator as class """

    def __init__(self, fcn: Callable):
        update_wrapper(self, fcn)
        self.calls_number = 0
        self.fcn = fcn

    def __call__(self, *args, **kwargs):
        self.calls_number += 1
        return self.fcn(*args, **kwargs)


def count_calls(fcn: Callable):
    """ Decorator as function """

    @wraps(fcn)
    def wrapper_count_calls(*args, **kwargs):
        wrapper_count_calls.calls_number += 1
        return fcn(*args, **kwargs)

    wrapper_count_calls.calls_number = 0
    return wrapper_count_calls



# -*- coding: utf-8 -*-

from multiprocessing.pool import Pool
from typing import Dict, List

from core_mixins.interfaces.task import ITask


class TasksManager:
    """ It manages the execution for the registered tasks """

    def __init__(self, tasks: List[ITask]):
        self.tasks = tasks

    def execute(self, task_name: str = None, parallelize: bool = False, processes: int = None) -> List[Dict]:
        """
        Execute all registered tasks. An exception in one task should not
        stop the execution of the others...

        :param task_name: If defined, only that specific task will be executed.
        :param parallelize: It defines if you want to execute the tasks in parallel.
        :param processes: Number of parallel process.

        :return: Result list in the form
            [
                {
                    "status": "Ok",
                    "result": ...
                },
                {
                    "status": "Failed",
                    "error": ...
                }
            ]
        """

        if task_name:
            for task in self.tasks:
                if task_name == task.name:
                    return [execute(task)]

        res = []
        if not parallelize:
            for task in self.tasks:
                res.append(execute(task))

        else:
            with Pool(processes=processes) as pool:
                res = pool.map(execute, (task for task in self.tasks))

        return res


def execute(task: ITask):
    try:
        return {
            "status": "Ok",
            "result": task.execute()
        }

    except Exception as error:
        return {
            "status": "Failed",
            "error": error
        }



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from inspect import isabstract
from typing import Type, Dict, Self, List


class IFactory(ABC):
    """ Base interface for classes that acts like factories """

    _impls: Dict[str, Type[Self]] = {}

    def __init_subclass__(cls):
        if not isabstract(cls):
            cls._impls[cls.registered_name()] = cls

    @classmethod
    def impls(cls) -> List[str]:
        """ It returns the current implementations """
        return [k for k in cls._impls.keys()]

    @classmethod
    @abstractmethod
    def registered_name(cls) -> str:
        """ It returns the name to use when the class is registered """

    @classmethod
    def get_implementation(cls, registered_name: str) -> Type[Self]:
        """ It returns the implementation class by the registered name """
        return cls._impls.get(registered_name, None)



# -*- coding: utf-8 -*-

from .cache import cache
from .count_calls import CountCalls, count_calls
from .repeat import repeat
from .retry import retry
from .singleton import singleton
from .timer import timer



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from enum import Enum
from logging import Logger
from typing import Optional, Dict, Type, Self

from core_mixins.interfaces.factory import IFactory


class TaskStatus(str, Enum):
    CREATED = "CREATED"
    EXECUTING = "EXECUTING"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


class ITask(IFactory, ABC):
    """ Base implementations for different tasks/processes """

    _impls: Dict[str, Type[Self]] = {}

    def __init__(
            self, name: Optional[str] = None, description: Optional[str] = None,
            logger: Optional[Logger] = None):

        self._name = name
        self.description = description
        self._status = TaskStatus.CREATED
        self.logger = logger

    @property
    def name(self):
        return self._name or self.__class__.__name__

    @property
    def status(self) -> TaskStatus:
        return self._status

    @status.setter
    def status(self, status: TaskStatus) -> None:
        self._status = status

    @abstractmethod
    def execute(self, *args, **kwargs):
        """ You must implement the task's process """

        self.info(f"Executing: {self.name}")
        self.info(f"Purpose: {self.description}")

    def info(self, message) -> None:
        if self.logger:
            self.logger.info(f"{self.name} | {message}")

    def warning(self, message) -> None:
        if self.logger:
            self.logger.warning(f"{self.name} | {message}")

    def error(self, error) -> None:
        if self.logger:
            self.logger.error(f"{self.name} | {error}")


class TaskException(Exception):
    """ Custom exception for Tasks """



# -*- coding: utf-8 -*-

from functools import wraps
from typing import Callable


def repeat(fcn: Callable = None, *, times: int = 2):
    """ Repeat n times the function and return the list of returned values """

    def decorator_repeat(func):
        @wraps(func)
        def wrapper_repeat(*args, **kwargs):
            values = []
            for _ in range(times):
                values.append(func(*args, **kwargs))

            return values

        return wrapper_repeat

    if not fcn:
        return decorator_repeat

    return decorator_repeat(fcn)



# -*- coding: utf-8 -*-

from functools import wraps
from logging import Logger
from time import sleep
from typing import Tuple, Callable


def retry(
        fcn: Callable = None, tries: int = 3, delay: int = 1, backoff: int = 2,
        exceptions: Tuple = (Exception,), logger: Logger = None):

    """ Retry calling the decorated function using an exponential backoff """

    def decorator_retry(_fcn):
        @wraps(_fcn)
        def function_retry(*args, **kwargs):
            _tries, _delay = tries, delay
            while _tries > 1:
                try:
                    return _fcn(*args, **kwargs)

                except exceptions as error:
                    if logger:
                        logger.warning(f"Retrying in {_delay}. Because: {error}")

                    sleep(delay)
                    _tries -= 1
                    _delay *= backoff

            return _fcn(*args, **kwargs)

        return function_retry

    if not fcn:
        return decorator_retry

    return decorator_retry(fcn)



# -*- coding: utf-8 -*-

import logging
import sys
import typing


def get_logger(
        logger_name: str = None, log_level: int = logging.INFO, propagate: bool = True,
        stream: typing.TextIO = sys.stdout, formatter: str = "[%(levelname)s] %(message)s",
        reset_handlers: bool = False) -> logging.Logger:

    """
    It returns a logger or create it if it doesn't exist. If the logger name
    is not specified, it returns the root logger....

    :param logger_name: Logger's name or None for root logger.
    :param log_level:  The logging level (CRITICAL = 50, ERROR = 40, WARNING = 30, INFO = 20)
    :param stream: A handler which writes logging records, appropriately formatted to a stream.

    :param formatter:
        It specifies the layout of log records in the final output. It defines the structure
        of log messages by specifying a format string that includes placeholders for
        various attributes of a log record, such as timestamp, log level,
        logger name, and the log message itself...

    :param propagate:
        It determines whether a log message should be passed to the handlers of
        higher-level (ancestor) loggers.This can be useful in situations where you want
        to control the flow of log messages and avoid duplication in the
        log output (like Lambda Functions)...

    :param reset_handlers:
        In case we want to remove all handlers and create a new one
        with the new stream...

    :return: The logger.
    """

    logger = logging.getLogger(logger_name)
    logger.propagate = propagate
    logger.setLevel(log_level)

    if reset_handlers:
        logger.handlers = []

    handler = logging.StreamHandler(stream)
    handler.setFormatter(logging.Formatter(formatter))
    handler.setLevel(log_level)
    logger.addHandler(handler)

    return logger



# -*- coding: utf-8 -*-

from functools import wraps
from time import perf_counter
from typing import Callable


def timer(fcn: Callable = None, precision: int = 4):
    def decorator_timer(func):
        @wraps(func)
        def wrapper_timer(*args, **kwargs):
            start_time, res = perf_counter(), func(*args, **kwargs)
            return res, f"{(perf_counter() - start_time):.{precision}f}"

        return wrapper_timer

    if not fcn:
        return decorator_timer

    return decorator_timer(fcn)



# -*- coding: utf-8 -*-

from functools import wraps


def singleton(cls):
    """ Make a class a Singleton class (only one instance) """

    @wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance:
            wrapper_singleton.instance = cls(*args, **kwargs)

        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton
